/*
 *   Copyright © 2011 ROSA Lab. <ural.mullabaev@rosalab.ru>
 *   Copyright © 2008, 2009 Fredrik Höglund <fredrik@kde.org>
 *   Copyright © 2008 Rafael Fernández López <ereslibre@kde.org>
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public License
 *   along with this library; see the file COPYING.LIB.  If not, write to
 *   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *   Boston, MA 02110-1301, USA.
 */

#include "stackfolder.h"

#include <QGraphicsLinearLayout>

#include <KDirModel>
#include <KDirLister>
#include <KFileItemDelegate>
#include <kfileplacesmodel.h>
#include <kfilepreviewgenerator.h>
#include <KGlobalSettings>
#include <KWindowSystem>

#include <limits.h>

#ifdef Q_OS_WIN
#  define _WIN32_WINNT 0x0500 // require NT 5.0 (win 2k pro)
#  include <windows.h>
#endif // Q_OS_WIN

#include <Plasma/WindowEffects>
#include <Plasma/IconWidget>
#include <Plasma/Theme>
#include <Plasma/ToolTipManager>

#include "stackfolderadapter.h"
#include "dialog.h"
#include "label.h"
#include "proxymodel.h"

K_EXPORT_PLASMA_APPLET(stackfolder, StackFolder)

StackFolder::StackFolder(QObject *parent, const QVariantList &args)
    : Plasma::Applet(parent, args),
      m_dialog(0),
      m_graphicsWidget(0),
      m_previewGenerator(0),
      m_placesModel(0),
      m_iconWidget(0),
      m_returnIcon(0),
      m_openIcon(0),
      m_label(0),
      m_iconView(0)
{
    setAspectRatioMode(Plasma::IgnoreAspectRatio);
    setHasConfigurationInterface(false);

    m_dirModel = new KDirModel(this);

    m_model = new ProxyModel(this);
    m_model->setSourceModel(m_dirModel);
    m_model->setSortLocaleAware(true);
    m_model->setFilterCaseSensitivity(Qt::CaseInsensitive);

    m_delegate = new KFileItemDelegate(this);
    m_selectionModel = new QItemSelectionModel(m_model, this);

    m_layout = 0;

    if (args.count() > 0) {
        setUrl(KUrl(args.value(0).toString()));
    }
}

void StackFolder::init()
{
    // Find out about icon and font settings changes
    connect(KGlobalSettings::self(), SIGNAL(kdisplayFontChanged()), SLOT(fontSettingsChanged()));
    connect(KGlobalSettings::self(), SIGNAL(iconChanged(int)), SLOT(iconSettingsChanged(int)));
    connect(/*KGlobalSettings::self()*/this, SIGNAL(geometryChanged()), SLOT(iconGeometryChanged()));

    // Find out about theme changes
    connect(Plasma::Theme::defaultTheme(), SIGNAL(themeChanged()), SLOT(themeChanged()));

    // Find out about network availability changes
    connect(Solid::Networking::notifier(), SIGNAL(statusChanged(Solid::Networking::Status)),
            SLOT(networkStatusChanged(Solid::Networking::Status)));

    m_customIconSize      = 64;
    m_showPreviews        = true;
    m_numTextLines        = 2;
    m_textColor           = QColor(Qt::transparent);
    m_iconsLocked         = false;
    m_alignToGrid         = false;
    m_previewPlugins      = QStringList() << "ffmpegthumbs" 		// Video
		  			  << "imagethumbnail" 		// Image
		  			  << "jpegthumbnail" 		// Jpeg
		  			  << "svgthumbnail" 		// Svg
		  			  << "windowsimagethumbnail"; 	// Windows images
		  			  // << "gsthumbnail" 		// PostScript, PDF, DVI
	  	  			  // << "djvuthumbnail" 	// DjVu
	  	  			  // << "textthumbnail" 	// Text
		  			  // << "opendocumentthumbnail" // ODF (OpenDocument Format)
		  			  // << "exrthumbnail" 		// EXR images
		  			  // << "comicbookthumbnail" 	// Comic books
		  			  // << "rawthumbnail" 		// RAW
		  			  // << "windowsexethumbnail" 	// Microsoft Windows exec
		  			  // << "desktopthumbnail" 	// Desktop
		  			  // << "fontthumbnail" 	// Fonts
		  			  // << "htmlthumbnail" 	// HTML
		  			  // << "mobithumbnail" 	// Mobipocket
		  			  // << "webarchivethumbnail" 	// Web archives
		  			  // << "directorythumbnail" 	// Directories
    m_sortDirsFirst       = true;
    m_sortColumn          = int(KDirModel::Name);
    m_filterType          = 0;
    m_hideForChangeFolder = true;


    m_flow = layoutDirection() == Qt::LeftToRight ? IconView::LeftToRight : IconView::RightToLeft;


    m_model->setFilterMode(ProxyModel::filterModeFromInt(m_filterType));
    m_model->setSortDirectoriesFirst(m_sortDirsFirst);
    m_model->setDynamicSortFilter(m_sortColumn != -1);
    m_model->sort(m_sortColumn != -1 ? m_sortColumn : KDirModel::Name, Qt::AscendingOrder);

    KDirLister *lister = new KDirLister(this);
    lister->setDelayedMimeTypes(true);
    lister->setAutoErrorHandlingEnabled(false, 0);

    m_dirModel->setDirLister(lister);
    connect(lister, SIGNAL(completed(const KUrl&)), SLOT(folderChanged(const KUrl&)));

    if (!m_url.isValid()) {
        QString path = QDir::homePath();
        setUrl(config().readEntry("url", KUrl(path)));
    } else {
        config().writeEntry("url", m_url);
    }

    lister->openUrl(m_url);

    m_topUrl = m_url;
    m_needShow = false;
}

StackFolder::~StackFolder()
{

}

void StackFolder::saveState(KConfigGroup &config) const
{
    Q_UNUSED(config)
    saveIconPositions();
}

QColor StackFolder::textColor() const
{
    if (m_textColor != Qt::transparent) {
        return m_textColor;
    }

    return Plasma::Theme::defaultTheme()->color(Plasma::Theme::TextColor);
}

void StackFolder::updateIconViewState()
{
    QPalette palette = m_iconView->palette();
    palette.setColor(QPalette::Text, textColor());
    m_iconView->setPalette(palette);

    m_iconView->setDrawShadows(m_drawShadows);
    m_iconView->setIconSize(iconSize());
    m_iconView->setTextLineCount(m_numTextLines);
    m_iconView->setFlow(m_flow);
    m_iconView->setWordWrap(m_numTextLines > 1);
    m_iconView->setAlignToGrid(m_alignToGrid);
    m_iconView->setIconsMoveable(!m_iconsLocked);

    if (m_label) {
        m_label->setPalette(palette);
        m_label->setDrawShadow(m_drawShadows);
    }
}

void StackFolder::constraintsEvent(Plasma::Constraints constraints)
{
    if (m_graphicsWidget) {
        return;
    }

    if (constraints & Plasma::FormFactorConstraint) {
        if (formFactor() != Plasma::Planar && formFactor() != Plasma::MediaCenter) {

            // Set up the icon widget
            m_iconWidget = new Plasma::IconWidget(KIcon("user-folder"), "", this);
            connect(m_iconWidget, SIGNAL(clicked()), SLOT(iconWidgetClicked()));

            //updateIconWidget();

	    QGraphicsLinearLayout *hlayout = new QGraphicsLinearLayout();
	    m_label = new Label(this);
	    m_label->setText(m_titleText);

	    QFont font = Plasma::Theme::defaultTheme()->font(Plasma::Theme::DesktopFont);
	    font.setPointSize(font.pointSize() + 1);
	    font.setBold(true);
	    m_label->setFont(font);
	    m_label->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed);

	    m_returnIcon = new Plasma::IconWidget(KIcon("go-previous"), "", this);
	    m_returnIcon->setPreferredSize(QSize(m_label->preferredHeight(), m_label->preferredHeight()));
	    m_returnIcon->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
	    m_returnIcon->setEnabled(false);

	    connect(m_returnIcon, SIGNAL(clicked()), this, SLOT(returnIconClicked()));

	    m_openIcon = new Plasma::IconWidget(KIcon("system-file-manager"), "", this);
	    m_openIcon->setPreferredSize(QSize(m_label->preferredHeight(), m_label->preferredHeight()));
	    m_openIcon->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);

	    connect(m_openIcon, SIGNAL(clicked()), this, SLOT(openIconClicked()));

	    hlayout->addItem(m_returnIcon);
	    hlayout->addItem(m_label);
	    hlayout->addItem(m_openIcon);

	    m_iconView = new IconView(this);
	    m_iconView->setModel(m_model);
	    m_iconView->setItemDelegate(m_delegate);
	    m_iconView->setSelectionModel(m_selectionModel);
	    m_iconView->setFont(Plasma::Theme::defaultTheme()->font(Plasma::Theme::DesktopFont));
	    m_iconView->setPreferredHeight(4 * m_label->preferredHeight());
	    m_iconView->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);

	    updateIconViewState();

	    const QStringList data = config().readEntry("savedPositions", QStringList());
	    m_iconView->setIconPositionsData(data);

	    connect(m_iconView, SIGNAL(activated(QModelIndex)), SLOT(activated(QModelIndex)));
	    connect(m_iconView, SIGNAL(indexesMoved(QModelIndexList)), SLOT(indexesMoved(QModelIndexList)));

	    StackFolderAdapter *adapter = new StackFolderAdapter(m_iconView);
	    m_previewGenerator = new KFilePreviewGenerator(adapter, m_model);
	    m_previewGenerator->setPreviewShown(m_showPreviews);
	    m_previewGenerator->setEnabledPlugins(m_previewPlugins);

	    m_layout = new QGraphicsLinearLayout(Qt::Vertical);
	    m_layout->addItem(hlayout);
	    m_layout->addItem(m_iconView);

	    m_graphicsWidget = new QGraphicsWidget(this);
	    m_graphicsWidget->setLayout(m_layout);

            m_dialog = new Dialog;
            m_dialog->setGraphicsWidget(m_graphicsWidget); // Ownership is transferred to the scene in the dialog

	    connect(m_dialog, SIGNAL(hidden()), this, SLOT(dialogHidden()));

            QGraphicsLinearLayout *layout = new QGraphicsLinearLayout(Qt::Vertical, this);
            layout->setContentsMargins(3, 0, 3, 6);
            layout->setSpacing(0);
            layout->addItem(m_iconWidget);

            setLayout(layout);
            setAspectRatioMode(Plasma::ConstrainedSquare);
        }
    }
}


void StackFolder::configChanged()
{
    const KUrl url = m_url;
    m_url = config().readEntry("url", m_url);
    if (url != m_url) {
        setUrl(m_url);
        m_dirModel->dirLister()->openUrl(m_url);
    }
}


void StackFolder::folderChanged(const KUrl& url)
{
    if (m_graphicsWidget && m_iconView) {
	const int spacing = 10;
	const int margin  = 10;
	const int count = m_dirModel->dirLister()->itemsForDir(url).count();

	QSize labelSize = m_label->size().toSize();
  	QSize viewSize = m_iconView->sizeToFitIcons(count);
  	QSize size = QSize(viewSize.width() + 2 * margin,
  			   labelSize.height() + viewSize.height() + 2 * margin + spacing);
	//kDebug() << " labelSize:" << labelSize << " oldViewSize:" << m_iconView->size().toSize() << " newViewSize:" << viewSize;
	//kDebug() << " size:" << size << " icons:"<< count;

  	m_graphicsWidget->resize(size);
  	m_graphicsWidget->setMinimumSize(size);
  	m_graphicsWidget->setMaximumSize(size);
	//kDebug() << " size after:" << m_graphicsWidget->size().toSize();
    }
    updateIconWidget();
}

void StackFolder::fontSettingsChanged()
{
    QFont font = Plasma::Theme::defaultTheme()->font(Plasma::Theme::DesktopFont);

    if (m_iconView) {
        m_iconView->setFont(font);
    }

    if (m_label) {
        font.setPointSize(font.pointSize() + 1);
        font.setBold(true);
        m_label->setFont(font);
    }
}

void StackFolder::iconSettingsChanged(int group)
{
    if (group == KIconLoader::Panel && m_iconView) {
        const int size = (m_customIconSize != 0) ?
                m_customIconSize : KIconLoader::global()->currentSize(KIconLoader::Panel);

        m_iconView->setIconSize(QSize(size, size));
        m_iconView->markAreaDirty(m_iconView->visibleArea());
        m_iconView->update();

    }
}

void StackFolder::iconGeometryChanged()
{
    if (m_iconWidget) {
        qDebug("StackFolder::iconGeometryChanged");
        m_iconWidget->update(m_iconWidget->boundingRect());
    }
}

void StackFolder::themeChanged()
{
    if (m_textColor != Qt::transparent) {
        return;
    }

    if (m_iconView) {
        QPalette palette = m_iconView->palette();
        palette.setColor(QPalette::Text, textColor());
        m_iconView->setPalette(palette);
    }

    if (m_label) {
        QPalette palette = m_label->palette();
        palette.setColor(QPalette::Text, textColor());
        m_label->setPalette(palette);
    }
}

void StackFolder::networkStatusChanged(Solid::Networking::Status status)
{
    if (status == Solid::Networking::Connected && !m_url.isLocalFile() &&
        m_url.protocol() != "desktop") {
        refreshIcons();
    }
}

void StackFolder::saveIconPositions() const
{
    if (!m_iconView) {
        return;
    }

    const QStringList data = m_iconView->iconPositionsData();
    if (!data.isEmpty()) {
        config().writeEntry("savedPositions", data);
    } else {
        config().deleteEntry("savedPositions");
    }
}

void StackFolder::setUrl(const KUrl &url)
{
    m_url = url;

    // Only parse desktop files when sorting if we're showing the desktop folder
    m_model->setParseDesktopFiles(m_url.protocol() == "desktop");
    m_titleText = m_url.fileName(KUrl::IgnoreTrailingSlash);
    if (m_label) {
        m_label->setText(m_titleText);
    }
}

void StackFolder::changeFolder(const KUrl &url, const bool toChild)
{
    setUrl(url);
    m_dirModel->dirLister()->openUrl(url/*, KDirLister::Reload*/);

    if (toChild) {
	if (m_topUrl.equals(url.upUrl(), KUrl::CompareWithoutTrailingSlash)&& m_returnIcon)
	    m_returnIcon->setEnabled(true);
    }
    else {
        if (m_topUrl.equals(url, KUrl::CompareWithoutTrailingSlash) && m_returnIcon)
	    m_returnIcon->setEnabled(false);
    }
}

void StackFolder::refreshIcons()
{
    m_dirModel->dirLister()->updateDirectory(m_url);
}

void StackFolder::updateIconWidget()
{
    if (!m_iconWidget) {
        return;
    }

    if (!m_placesModel) {
	m_placesModel = new KFilePlacesModel(this);

        const QModelIndex index = m_placesModel->closestItem(m_topUrl/*m_url*/);
	const KUrl url = m_placesModel->url(index);

        KIcon icon;
	KFileItem item = m_dirModel->itemForIndex(QModelIndex());
        if (!item.isNull() && item.iconName() != "inode-directory") {
	    icon = KIcon(item.iconName(), 0, item.overlays());
        } else if (m_topUrl.protocol() == "desktop") {
	    icon = KIcon("user-desktop");
        } else if (m_topUrl.protocol() == "trash") {
	    icon = m_model->rowCount() > 0 ? KIcon("user-trash-full") : KIcon("user-trash");
        } else if (index.isValid() && url.equals(m_topUrl, KUrl::CompareWithoutTrailingSlash)) {
	    icon = m_placesModel->icon(index);
        } else {
	    icon = KIcon("folder");
        }
        m_iconWidget->setIcon(icon);
	m_iconWidget->update();
    }

    // Update the tooltip
    Plasma::ToolTipContent data;
    data.setMainText(i18n("Stack Folder"));
    data.setSubText(m_titleText);
    data.setImage(m_iconWidget->icon());
    Plasma::ToolTipManager::self()->setContent(m_iconWidget, data);
}

void StackFolder::iconWidgetClicked()
{
    Plasma::WindowEffects::slideWindow(m_dialog, location());
    if (m_dialog->isVisible()) {
        hideDialog();
    } else {
        showDialog();
    }
}

void StackFolder::returnIconClicked()
{
    if (m_topUrl.isParentOf(m_url) &&
	!m_topUrl.equals(m_url, KUrl::CompareWithoutTrailingSlash)) {

	changeFolder(m_url.upUrl(), false);
    }

    if (m_hideForChangeFolder) {
	m_needShow = true;
	hideDialog();
    }
}

void StackFolder::openIconClicked()
{
    const QModelIndex index = m_dirModel->indexForUrl(m_url);
    const KFileItem item = m_dirModel->itemForIndex(index);
    item.run();
    hideDialog();
}

void StackFolder::activated(const QModelIndex &index)
{
    const KFileItem item = m_model->itemForIndex(index);
    if (item.isDir()) {

	changeFolder(item.url(), true);

	if (m_hideForChangeFolder) {
	    m_needShow = true;
	    hideDialog();
	}
    }
    else if (item.isFile()) {
  	item.run();
	hideDialog();
    }

    m_returnIcon->setEnabled(true);
}

void StackFolder::indexesMoved(const QModelIndexList &indexes)
{
    Q_UNUSED(indexes)

    // If the user has rearranged the icons, the view is no longer sorted
    if (m_sortColumn != -1) {
        m_sortColumn = -1;
        config().writeEntry("sortColumn", m_sortColumn);
        emit configNeedsSaving();
    }

    m_delayedSaveTimer.start(5000, this);
}

void StackFolder::dialogHidden()
{
    if (m_needShow) {
	if (KWindowSystem::compositingActive())
    	    m_delayedShowTimer.start(200, this);
	else
	    m_delayedShowTimer.start(100, this);
	m_needShow = false;
    }
    else {
	if (!m_topUrl.equals(m_url, KUrl::CompareWithoutTrailingSlash)) {
	    changeFolder(m_topUrl, false);
	}
    }
}

void StackFolder::showDialog()
{
    Plasma::WindowEffects::slideWindow(m_dialog, location());
    m_dialog->show(this);
}

void StackFolder::hideDialog()
{
    Plasma::WindowEffects::slideWindow(m_dialog, location());
    m_dialog->hide();
}

QSize StackFolder::iconSize() const
{
    const int defaultSize = KIconLoader::global()->currentSize( KIconLoader::Panel);
    const int size = (m_customIconSize != 0) ? m_customIconSize : defaultSize;
    return QSize(size, size);
}

bool StackFolder::eventFilter(QObject *watched, QEvent *event)
{
    if (event->type() == QEvent::ContextMenu) {
        return false;
    }
    return Applet::eventFilter(watched, event);
}

void StackFolder::timerEvent(QTimerEvent *event)
{
    if (event->timerId() == m_delayedShowTimer.timerId()) {
        m_delayedShowTimer.stop();
	showDialog();
    }
    if (event->timerId() == m_delayedSaveTimer.timerId()) {
        m_delayedSaveTimer.stop();
        saveIconPositions();
        emit configNeedsSaving();
    }
    Applet::timerEvent(event);
}

QSizeF StackFolder::sizeHint(Qt::SizeHint which, const QSizeF &constraint) const
{

    if (which == Qt::PreferredSize) {
        if (!constraint.isEmpty()) {
            return QSizeF(400, 400).boundedTo(constraint);
        } else {
            return QSizeF(400, 400);
        }
    }

    return Applet::sizeHint(which, constraint);
}

#include "stackfolder.moc"